#ifndef TEXTFILEREADER_H
#define TEXTFILEREADER_H

#include <fstream>
#include <iomanip>
#include <iostream>
#include <typeinfo>
#include <vector>

/**
 
  \brief Text file reader
  \details class to read text file like input parameters or data tables 
  \author Wouf (Wouf@mail.cern.ch)
  \version 1.1
  \date Nov 2019
  \copyright GNU Public License.
*/

class TextFileReader {

private:
  /**
 \details Table key, or mask: is table struct AFTER KEYWORD PARAMETER looks like "2i|3d|f|2i" or "d:i:i:i" or "2F;4d;1f" where i(nt), f(loat), d(ouble) is the types, numbers is multiplyers ("i|i|i" is the same as "3i") and the symbols "|", ":", ";" or " " - is column separators
*/
  std::string tableKey;
  std::ifstream TextFile; //< Text file name parameter
  const char *fileToRead;

  void SetSeparator() {
    /// Care about float separator, seting to "."
    char *current = setlocale(LC_ALL, NULL);
    std::locale::global(std::locale(std::locale(current), "C", std::locale::numeric));
  }

public:
  std::vector<int> intParamValue;       //< \param [out] intParamValue Int output vector
  std::vector<float> floatParamValue;   //< \param [out] floatParamValue Float output vector
  std::vector<double> doubleParamValue; //< \param [out] doubleParamValue Double output vector

  TextFileReader(const char *FileToRead, std::string TableKey = "") : tableKey(TableKey), fileToRead(FileToRead) {
    SetSeparator();
    TextFile.open(FileToRead, std::ios::in);
    if (!TextFile.good()) {
      std::clog << "Error!!! TextFileReader: File " << FileToRead << " read error!" << std::endl;
      exit(1);
    }
  }

  ~TextFileReader() {
    TextFile.close();
    intParamValue.clear();
    floatParamValue.clear();
    doubleParamValue.clear();
  }

  /// Type ReadParam<Type>, where the type is int, float or double, result is returned
  /// \param [in] ParamToFind String keyword to look for
  /// \return int, float or double
  template <typename P>
  P ReadParam(std::string ParamToFind) {
    for (short attempt = 1; attempt < 3; attempt++) {
      std::string ReadedWord;
      while (TextFile >> ReadedWord)
        if (ReadedWord == ParamToFind) {
          // next to parameter name value
          TextFile >> ReadedWord;
          if ((ReadedWord == "=") || (ReadedWord == "-") || (ReadedWord == "is") || (ReadedWord == ":"))
            TextFile >> ReadedWord;

          P ParamValue;
          std::size_t ReadedWordSize = 0;
          if (typeid(ParamValue) == typeid(int)) {
            ParamValue = std::stoi(ReadedWord, &ReadedWordSize);
            std::size_t ReadedWordSizeD = 0;
            std::stod(ReadedWord, &ReadedWordSizeD);
            if (ReadedWordSize < ReadedWordSizeD) {
              std::clog << "Error!!! TextFileReader: Attempt to read DOUBLE as INT, at parameter " << ParamToFind << " = " << ReadedWord << " in file " << fileToRead << std::endl;
              exit(1);
            }
          } else if (typeid(ParamValue) == typeid(float))
            ParamValue = std::stof(ReadedWord, &ReadedWordSize);
          else if (typeid(ParamValue) == typeid(double))
            ParamValue = std::stod(ReadedWord, &ReadedWordSize);
          else {
            std::clog << "Error!!! TextFileReader: parameter " << ParamToFind << " from input file " << fileToRead << " has wrong format" << std::endl;
            exit(1);
          }

          if (attempt > 1)
            std::clog << "Warning!!! TextFileReader: parameter's sequence in file " << fileToRead << " and code are different" << std::endl;
          return ParamValue;
        }

      if (ReadedWord != ParamToFind) {
        if (attempt == 1) {
          TextFile.clear();
          TextFile.seekg(0, std::ios::beg);
        } else {
          std::clog << "Error!!! TextFileReader: parameter " << ParamToFind << " not found in input file " << fileToRead << std::endl;
          exit(1);
        }
      }
    }

    std::clog << "Error!!! TextFileReader: ReadParam error at parameter " << ParamToFind << " in file " << fileToRead << "!!! Please take note to the autors!!!" << std::endl;
    exit(1);
  }

  /**
 \details Function to find string by int number. Result will be stored in 3 vectors: 
intParamValue Int output vector
floatParamValue Float output vector
doubleParamValue Double output vector
Results for mask looks like "2i|1d|1i", would be "intParamValue.at(0) | intParamValue.at(1) |  doubleParamValue.at(0) | intParamValue.at(2)"
\param [in] NumberToFind Digit to find 
*/

  void ReadTable(int NumberToFind) {
    std::string ParamToFind = std::to_string(NumberToFind);
    ReadTable(ParamToFind);
  }
  void ReadTable(float NumberToFind) {
    std::clog << "Error!!! TextFileReader: ReadTable error at parameter " << NumberToFind << ". Using sarch by FLOAT parameter is not possible!!!" << std::endl;
    exit(1);
  }
  void ReadTable(double NumberToFind) {
    std::clog << "Error!!! TextFileReader: ReadTable error at parameter " << NumberToFind << ". Using sarch by DOUBLE parameter is not possible!!!" << std::endl;
    exit(1);
  }

  void ReadTable(const char *paramToFind) {
    std::string ParamToFind(paramToFind);
    ReadTable(ParamToFind);
  }
  /**
 \details Function to find string by text parameter name. Result will be stored in 3 vectors: 
intParamValue Int output vector
floatParamValue Float output vector
doubleParamValue Double output vector
Results for mask looks like "2i|1d|1i", would be "intParamValue.at(0) | intParamValue.at(1) |  doubleParamValue.at(0) | intParamValue.at(2)"
\param [in] ParamToFind Keyword to find 
*/
  void ReadTable(std::string ParamToFind) {
    if (tableKey.size() == 0) {
      std::clog << "Error!!! TextFileReader: please provide table mask befor calling ReadTable(...)" << std::endl;
      exit(1);
    }

    intParamValue.clear();
    floatParamValue.clear();
    doubleParamValue.clear();

    for (short attempt = 1; attempt < 3; attempt++) {
      std::string ReadedWord;
      while (TextFile >> ReadedWord)
        if (ReadedWord == ParamToFind) {
          std::getline(TextFile, ReadedWord);

          short position = 0, nColumn = 1;
          std::size_t ReadedWordSize = 0;
          for (auto &column : tableKey) {
            if ((column == 'i') || (column == 'I')) {
              for (short i = 1; i <= nColumn; i++) {
                intParamValue.push_back(std::stoi(&ReadedWord[position += ReadedWordSize], &ReadedWordSize));
                std::size_t ReadedWordSizeD = 0;
                std::stod(static_cast<const std::string &>(&ReadedWord[position]), &ReadedWordSizeD);
                if (ReadedWordSize < ReadedWordSizeD) {
                  std::clog << "Error!!! TextFileReader: Attempt to read DOUBLE as INT, at parameter " << ParamToFind << " = " << std::stod(static_cast<const std::string &>(&ReadedWord[position])) << " in file " << fileToRead << std::endl;
                  exit(1);
                }
              }
            } else if ((column == 'f') || (column == 'F')) {
              for (short i = 1; i <= nColumn; i++)
                floatParamValue.push_back(std::stof(&ReadedWord[position += ReadedWordSize], &ReadedWordSize));
            } else if ((column == 'd') || (column == 'D')) {
              for (short i = 1; i <= nColumn; i++)
                doubleParamValue.push_back(std::stod(static_cast<const std::string &>(&ReadedWord[position += ReadedWordSize]), &ReadedWordSize));
            } else if ((48 <= int(column)) && (int(column) <= 57)) {
              nColumn = int(column) - 48;
              continue;
            } else if ((column == ':') || (column == '|') || (column == ';') || (column == ' ') || (column == ',')) {
              nColumn = 1;
              continue;
            }
          }
          if (attempt > 1)
            std::clog << "Warning!!! TextFileReader: parameter's sequence in file " << fileToRead << " and code are different" << std::endl;
          return;
        }
      if (ReadedWord != ParamToFind) {
        if (attempt == 1) {
          TextFile.clear();
          TextFile.seekg(0, std::ios::beg);
        } else {
          std::clog << "Error!!! TextFileReader: parameter " << ParamToFind << " not found in input file " << fileToRead << std::endl;
          exit(1);
        }
      }
    }

    std::clog << "Error!!! TextFileReader: ReadTable error at parameter " << ParamToFind << " in file " << fileToRead << "!!! Please take note to the autors!!!" << std::endl;
    exit(1);
  }
};

#endif
